@@ -44,6 +44,7 @@ struct connman_config_service {
 	char *private_key_passphrase;
 	char *private_key_passphrase_type;
 	char *phase2;
+	char *passphrase;
 };
 
 struct connman_config {
@@ -55,6 +56,45 @@ struct connman_config {
 
 static GHashTable *config_table = NULL;
 
+/* Definition of possible strings in the .config files */
+#define CONFIG_KEY_NAME                "Name"
+#define CONFIG_KEY_DESC                "Description"
+
+#define SERVICE_KEY_TYPE               "Type"
+#define SERVICE_KEY_NAME               "Name"
+#define SERVICE_KEY_SSID               "SSID"
+#define SERVICE_KEY_EAP                "EAP"
+#define SERVICE_KEY_CA_CERT            "CACertFile"
+#define SERVICE_KEY_CL_CERT            "ClientCertFile"
+#define SERVICE_KEY_PRV_KEY            "PrivateKeyFile"
+#define SERVICE_KEY_PRV_KEY_PASS       "PrivateKeyPassphrase"
+#define SERVICE_KEY_PRV_KEY_PASS_TYPE  "PrivateKeyPassphraseType"
+#define SERVICE_KEY_IDENTITY           "Identity"
+#define SERVICE_KEY_PHASE2             "Phase2"
+#define SERVICE_KEY_PASSPHRASE         "Passphrase"
+
+static const char *config_possible_keys[] = {
+	CONFIG_KEY_NAME,
+	CONFIG_KEY_DESC,
+	NULL,
+};
+
+static const char *service_possible_keys[] = {
+	SERVICE_KEY_TYPE,
+	SERVICE_KEY_NAME,
+	SERVICE_KEY_SSID,
+	SERVICE_KEY_EAP,
+	SERVICE_KEY_CA_CERT,
+	SERVICE_KEY_CL_CERT,
+	SERVICE_KEY_PRV_KEY,
+	SERVICE_KEY_PRV_KEY_PASS,
+	SERVICE_KEY_PRV_KEY_PASS_TYPE,
+	SERVICE_KEY_IDENTITY,
+	SERVICE_KEY_PHASE2,
+	SERVICE_KEY_PASSPHRASE,
+	NULL,
+};
+
 static void unregister_config(gpointer data)
 {
 	struct connman_config *config = data;
@@ -87,9 +127,37 @@ static void unregister_service(gpointer data)
 	g_free(service->private_key_passphrase);
 	g_free(service->private_key_passphrase_type);
 	g_free(service->phase2);
+	g_free(service->passphrase);
 	g_free(service);
 }
 
+static void check_keys(GKeyFile *keyfile, const char *group,
+			const char **possible_keys)
+{
+	char **avail_keys;
+	gsize nb_avail_keys, i, j;
+
+	avail_keys = g_key_file_get_keys(keyfile, group, &nb_avail_keys, NULL);
+	if (avail_keys == NULL)
+		return;
+
+	/*
+	 * For each key in the configuration file,
+	 * verify it is understood by connman
+	 */
+	for (i = 0 ; i < nb_avail_keys; i++) {
+		for (j = 0; possible_keys[j] ; j++)
+			if (g_strcmp0(avail_keys[i], possible_keys[j]) == 0)
+				break;
+
+		if (possible_keys[j] == NULL)
+			connman_warn("Unknown configuration key %s in [%s]",
+					avail_keys[i], group);
+	}
+
+	g_strfreev(avail_keys);
+}
+
 static int load_service(GKeyFile *keyfile, const char *group,
 						struct connman_config *config)
 {
@@ -103,6 +171,9 @@ static int load_service(GKeyFile *keyfile, const char *group,
 	if (strlen(ident) < 1)
 		return -EINVAL;
 
+	/* Verify that provided keys are good */
+	check_keys(keyfile, group, service_possible_keys);
+
 	service = g_hash_table_lookup(config->service_table, ident);
 	if (service == NULL) {
 		service = g_try_new0(struct connman_config_service, 1);
@@ -112,19 +183,20 @@ static int load_service(GKeyFile *keyfile, const char *group,
 		service->ident = g_strdup(ident);
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Type", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_TYPE, NULL);
 	if (str != NULL) {
 		g_free(service->type);
 		service->type = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Name", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_NAME, NULL);
 	if (str != NULL) {
 		g_free(service->name);
 		service->name = str;
 	}
 
-	hex_ssid = g_key_file_get_string(keyfile, group, "SSID", NULL);
+	hex_ssid = g_key_file_get_string(keyfile, group, SERVICE_KEY_SSID,
+					 NULL);
 	if (hex_ssid != NULL) {
 		char *ssid;
 		unsigned int i, j = 0, hex;
@@ -161,56 +233,63 @@ static int load_service(GKeyFile *keyfile, const char *group,
 		service->ssid_len = ssid_len;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "EAP", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_EAP, NULL);
 	if (str != NULL) {
 		g_free(service->eap);
 		service->eap = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "CACertFile", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_CA_CERT, NULL);
 	if (str != NULL) {
 		g_free(service->ca_cert_file);
 		service->ca_cert_file = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "ClientCertFile", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_CL_CERT, NULL);
 	if (str != NULL) {
 		g_free(service->client_cert_file);
 		service->client_cert_file = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "PrivateKeyFile", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_PRV_KEY, NULL);
 	if (str != NULL) {
 		g_free(service->private_key_file);
 		service->private_key_file = str;
 	}
 
 	str = g_key_file_get_string(keyfile, group,
-						"PrivateKeyPassphrase", NULL);
+						SERVICE_KEY_PRV_KEY_PASS, NULL);
 	if (str != NULL) {
 		g_free(service->private_key_passphrase);
 		service->private_key_passphrase = str;
 	}
 
 	str = g_key_file_get_string(keyfile, group,
-					"PrivateKeyPassphraseType", NULL);
+					SERVICE_KEY_PRV_KEY_PASS_TYPE, NULL);
 	if (str != NULL) {
 		g_free(service->private_key_passphrase_type);
 		service->private_key_passphrase_type = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Identity", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_IDENTITY, NULL);
 	if (str != NULL) {
 		g_free(service->identity);
 		service->identity = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Phase2", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_PHASE2, NULL);
 	if (str != NULL) {
 		g_free(service->phase2);
 		service->phase2 = str;
 	}
 
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_PASSPHRASE,
+					NULL);
+	if (str != NULL) {
+		g_free(service->passphrase);
+		service->passphrase = str;
+	}
+
 	g_hash_table_replace(config->service_table, service->ident, service);
 
 	connman_info("Adding service configuration %s", service->ident);
@@ -232,13 +311,16 @@ static int load_config(struct connman_config *config)
 	if (keyfile == NULL)
 		return -EIO;
 
-	str = g_key_file_get_string(keyfile, "global", "Name", NULL);
+	/* Verify keys validity of the global section */
+	check_keys(keyfile, "global", config_possible_keys);
+
+	str = g_key_file_get_string(keyfile, "global", CONFIG_KEY_NAME, NULL);
 	if (str != NULL) {
 		g_free(config->name);
 		config->name = str;
 	}
 
-	str = g_key_file_get_string(keyfile, "global", "Description", NULL);
+	str = g_key_file_get_string(keyfile, "global", CONFIG_KEY_DESC, NULL);
 	if (str != NULL) {
 		g_free(config->description);
 		config->description = str;
@@ -440,6 +522,9 @@ static void provision_service(gpointer key, gpointer value, gpointer user_data)
 
 	if (config->phase2 != NULL)
 		__connman_service_set_string(service, "Phase2", config->phase2);
+
+	if (config->passphrase != NULL)
+		__connman_service_set_string(service, "Passphrase", config->passphrase);
 }
 
 int __connman_config_provision_service(struct connman_service *service)
