@@ -165,7 +165,8 @@ int __connman_provider_disconnect(struct connman_provider *provider)
 	else
 		return -EOPNOTSUPP;
 
-	__connman_service_indicate_state(provider->vpn_service,
+	if (provider->vpn_service != NULL)
+		__connman_service_indicate_state(provider->vpn_service,
 					CONNMAN_SERVICE_STATE_DISCONNECT);
 	if (err < 0) {
 		if (err != -EINPROGRESS)
@@ -215,18 +216,30 @@ int __connman_provider_connect(struct connman_provider *provider)
 int __connman_provider_remove(const char *path)
 {
 	struct connman_provider *provider;
+	GHashTableIter iter;
+	gpointer value, key;
 
 	DBG("path %s", path);
 
-	provider = g_hash_table_lookup(provider_hash, path);
-	if (provider == NULL) {
-		DBG("patch %s not found", path);
-		return -ENXIO;
+	g_hash_table_iter_init(&iter, provider_hash);
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		const char *srv_path;
+		provider = value;
+
+		if (provider->vpn_service == NULL)
+			continue;
+
+		srv_path = __connman_service_get_path(provider->vpn_service);
+
+		if (g_strcmp0(srv_path, path) == 0) {
+			DBG("Removing VPN %s", provider->identifier);
+			g_hash_table_remove(provider_hash,
+						provider->identifier);
+			return 0;
+		}
 	}
 
-	g_hash_table_remove(provider_hash, path);
-
-	return 0;
+	return -ENXIO;
 }
 
 static int set_connected(struct connman_provider *provider,
@@ -333,10 +346,12 @@ int connman_provider_set_state(struct connman_provider *provider,
 static void unregister_provider(gpointer data)
 {
 	struct connman_provider *provider = data;
+	struct connman_service *service = provider->vpn_service;
 
 	DBG("provider %p", provider);
 
-	__connman_service_put(provider->vpn_service);
+	provider->vpn_service = NULL;
+	__connman_service_put(service);
 
 	connman_element_unregister(&provider->element);
 	connman_provider_unref(provider);
@@ -419,9 +434,15 @@ static void provider_dbus_ident(char *ident)
 {
 	int i, len = strlen(ident);
 
-	for (i = 0; i < len; i++)
-		if (ident[i] == '.')
-			ident[i] = '_';
+	for (i = 0; i < len; i++) {
+		if (ident[i] >= '0' && ident[i] <= '9')
+			continue;
+		if (ident[i] >= 'a' && ident[i] <= 'z')
+			continue;
+		if (ident[i] >= 'A' && ident[i] <= 'Z')
+			continue;
+		ident[i] = '_';
+	}
 }
 
 int __connman_provider_create_and_connect(DBusMessage *msg)
@@ -531,7 +552,8 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 	if (created == TRUE)
 		provider_probe(provider);
 
-	provider->vpn_service =
+	if (provider->vpn_service == NULL)
+		provider->vpn_service =
 			__connman_service_create_from_provider(provider);
 	if (provider->vpn_service == NULL) {
 		err = -EOPNOTSUPP;
@@ -553,8 +575,10 @@ failed:
 		DBG("can not connect delete provider");
 		connman_provider_unref(provider);
 
-		if (provider->vpn_service != NULL)
+		if (provider->vpn_service != NULL) {
 			__connman_service_put(provider->vpn_service);
+			provider->vpn_service = NULL;
+		}
 	}
 
 	return err;
@@ -634,6 +658,9 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 
 	DBG("");
 
+	if (service == NULL)
+		return;
+
 	ipconfig = __connman_service_get_ipconfig(service);
 
 	if (ipconfig == NULL) {
@@ -688,12 +715,12 @@ void connman_provider_driver_unregister(struct connman_provider_driver *driver)
 	driver_list = g_slist_remove(driver_list, driver);
 }
 
-static void provider_disconnect(gpointer key, gpointer value,
+static void provider_remove(gpointer key, gpointer value,
 						gpointer user_data)
 {
 	struct connman_provider *provider = value;
 
-	__connman_provider_disconnect(provider);
+	g_hash_table_remove(provider_hash, provider->identifier);
 }
 
 static void provider_offline_mode(connman_bool_t enabled)
@@ -701,7 +728,7 @@ static void provider_offline_mode(connman_bool_t enabled)
 	DBG("enabled %d", enabled);
 
 	if (enabled == TRUE)
-		g_hash_table_foreach(provider_hash, provider_disconnect, NULL);
+		g_hash_table_foreach(provider_hash, provider_remove, NULL);
 
 }
 
